package vpc.cfg;

import vpc.tir.*;
import vpc.tir.expr.Operator;
import vpc.util.Pair;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Arrays;
import java.util.Collections;

public class Cfg
{
    public final Node[] nodes;
        // The ordering of this array is roughly program order.  So iterative
        // dataflow-type algorithms can use this order directly.

    public final Node entry;
        // The first node.  The first node has zero incoming edges.

    public Cfg(Node[] nodes, Node entry)
    {
        this.nodes = nodes;
        this.entry = entry;
    }

    public static Cfg build(TIRExpr tir)
    {
        NodeFactory factory = new NodeFactory();

        Node entry = factory.mk(); // Create an extra node to ensure the entry node has no incoming edges.
        Node next = factory.mk();
        entry.setTrans(new Trans.Goto(entry, next));

        Node exit = build(factory, tir, next, new IdentityHashMap<TIRBlock, BlockMeta>());
        if (exit != null) {
            exit.setTrans(new Trans.Return(null));
        }

        List<Node> all = factory.getAll();
        //DEBUG: CfgDotPrinter.print(new java.io.PrintWriter(System.err, true), "Before", new Cfg(all.toArray(new Node[0]), entry), CfgNodeRenderer.Instance);
        check(all, entry, newTrueArray(all.size())); // Check after construction.

        simplify(all, entry);

        List<Node> reachable = fixup(all, entry);

        Cfg cfg = new Cfg(reachable.toArray(new Node[reachable.size()]), entry);

        return cfg;
    }

    private static boolean[] newTrueArray(int size)
    {
        boolean[] a = new boolean[size];
        Arrays.fill(a, true);
        return a;
    }

    // Sanity check.  Will assert fail if the graph is not well formed.
    public static void check(List<Node> nodes, Node entry, boolean[] reachable)
    {
        assert entry == nodes.get(0);
        assert reachable[0];

        for (Node n : nodes) {
            if (!reachable[n.index]) continue;

            assert n.trans != null;

            // Check outgoing edges.
            for (Edge e : n.trans.getEdges()) {
                assert e.getTrans() == n.trans;
                assert reachable[e.target.index] : "Bad Trans: " + n.trans + ": Unreachable Target: " + e.target.index;
            }

            // Check incoming edges.
            for (Edge e : n.incoming) {
                Node edgeSource = e.getTrans().source;
                assert reachable[edgeSource.index];
                assert e.target == n;
            }
        }

        // TODO: Make sure the Trans entries match up with the corresponding
        // RTrans entries.  Currently, we only check that the incoming/outgoing
        // edges are balanced in number.  We don't check the transition type.
    }

    // Simplify the graph by removing useless nodes and transitions.
    private static void simplify(List<Node> all, Node entry)
    {
        boolean[] removed = new boolean[all.size()];

        // Remove useless transitions and useless blocks.
        {
            Iterator<Node> it = all.iterator();
            while (it.hasNext()) {
                Node n = it.next();
                if (removed[n.index]) continue;

                // If there's an edge where: sourceNode.outgoingCount=1 and targetNode.incomingCount=1,
                // then it's a useless edge.  Merge the two nodes.
                if (n.trans instanceof Trans.Goto) {
                    Node target = ((Trans.Goto) n.trans).edge.target;
                    if (target.incoming.size() == 1) {
                        // 1. Merge 'target' into 'n'.
                        n.stmts.addAll(target.stmts);
                        // 2. Forward 'n's transition all the way through 'target's transition.
                        Trans t = target.trans;
                        target.clearTrans();
                        n.clearTrans();
                        n.setTrans(t.changeSource(n));
                        removed[target.index] = true;
                    }
                }

                // Empty block with one outgoing edge.  Remove block and reroute incoming edges to
                // their final destination.
                if (n.trans instanceof Trans.Goto) {
                    Node target = ((Trans.Goto) n.trans).edge.target;
                    if (n.stmts.isEmpty()) {
                        // Don't do the transform for the entry node.  We want to make sure the
                        // entry node is preserved and that it has no incoming edges.
                        if (n != entry) {
                            // Get rid of the empty node and link each predecessor of 'n' directly
                            // to target.
                            n.clearTrans();
                            moveIncoming(n, target);
                            removed[n.index] = true;
                        }
                    }
                }
            }
        }

        assert !removed[0];
        assert entry.incoming.isEmpty();
    }

    private static void moveIncoming(Node from, Node to)
    {
        Trans newT;
        List<Edge> incoming = new ArrayList<Edge>(from.incoming);

        for (Edge ed : incoming) {
            if (ed instanceof Edge.Branch) {
                Edge.Branch rb = (Edge.Branch) ed;
                Trans.Branch oldT = rb.getTrans();
                newT = oldT.changeTarget(rb, to);
            }
            else if (ed instanceof Edge.Goto) {
                Edge.Goto rg = (Edge.Goto) ed;
                Trans.Goto oldT = rg.getTrans();
                newT = oldT.changeTarget(to);
            }
            else if (ed instanceof Edge.Switch.Case) {
                Edge.Switch.Case c = (Edge.Switch.Case) ed;
                Trans.Switch oldT = c.getTrans();
                newT = oldT.changeCaseTarget(c, to);
            }
            else if (ed instanceof Edge.Switch.Default) {
                Edge.Switch.Default def = (Edge.Switch.Default) ed;
                Trans.Switch oldT = def.getTrans();
                newT = oldT.changeDefaultTarget(to);
            }
            else {
                throw new AssertionError("bad type: " + ed.getClass().getName());
            }
            Node source = ed.getTrans().source;
            source.clearTrans();
            source.setTrans(newT);
        }
    }

    private static List<Node> fixup(List<Node> all, Node entry)
    {
        // Find reachable nodes.
        boolean[] reachable = new boolean[all.size()];
        markReachable(entry, reachable);

        // Make sure graph is consistent.
        check(all, entry, reachable);

        // Save only the reachable nodes into 'result'
        ArrayList<Node> result = new ArrayList<Node>();
        for (Node n : all) {
            if (reachable[n.index]) {
                result.add(n);
            }
        }

        //DEBUG: CfgDotPrinter.print(new java.io.PrintWriter(System.err, true), "After", new Cfg(result.toArray(new Node[0]), entry), CfgNodeRenderer.Instance);
        
        // Renumber nodes and edges
        int currentIndex = 0;
        for (Node n : result) {
            n.index = currentIndex++;
            int ei = 0;
            for (Edge e : n.incoming) {
                e.setIndex(ei++);
            }
        }

        return result;
    }

    private static void markReachable(Node n, boolean[] reachable)
    {
        if (reachable[n.index]) return;
        reachable[n.index] = true;

        Trans t = n.trans;
        assert t != null : n.index;
        if (t instanceof Trans.Goto) {
            Trans.Goto g = (Trans.Goto)t;
            markReachable(g.edge.target, reachable);
        }
        else if (t instanceof Trans.Branch) {
            Trans.Branch b = (Trans.Branch)t;
            markReachable(b.trueEdge.target, reachable);
            markReachable(b.falseEdge.target, reachable);
        }
        else if (t instanceof Trans.Switch) {
            Trans.Switch s = (Trans.Switch)t;
            for (Trans.Switch.Case c : s.cases) {
                markReachable(c.edge.target, reachable);
            }
            if (s.defEdge != null) {
                markReachable(s.defEdge.target, reachable);
            }
        }
        else if (t instanceof Trans.Return || t instanceof Trans.Throw) {
            // No edges here.
        }
        else {
            throw new AssertionError("bad type: " + t.getClass().getName());
        }
    }

    private static final class NodeFactory
    {
        private final List<Node> nodes = new ArrayList<Node>();

        public Node mk()
        {
            Node n = new Node(nodes.size());
            nodes.add(n);
            return n;
        }

        public List<Node> getAll() { return nodes; }
    }


    private static final class BlockMeta
    {
        public final Node continueTarget;
        public final ArrayList<Node> breakSources;

        private BlockMeta(Node continueTarget, ArrayList<Node> breakSources)
        {
            this.continueTarget = continueTarget;
            this.breakSources = breakSources;
        }
    }

    private static Node build(NodeFactory factory, TIRExpr tir, Node entry, Map<TIRBlock, BlockMeta> blockMap)
    {
        if (tir instanceof TIRIfExpr) {
            TIRIfExpr i = (TIRIfExpr) tir;

            Node tHead = factory.mk();
            Node tExit = build(factory, i.true_target, tHead, blockMap);

            Node exit;

            if (i.false_target == null) {
                // There's only a "true" branch (no "else" branch)
                exit = factory.mk();

                // entry node --> (true branch, end node)
                entry.setTrans(new Trans.Branch(tExit, i.condition, tHead, exit));

                // true branch --> end node
                if (tExit != null) {
                    tExit.setTrans(new Trans.Goto(tExit, exit));
                }
            }
            else {
                // There are both "true" and "false" branches.
                Node fHead = factory.mk();
                Node fExit = build(factory, i.false_target, fHead, blockMap);

                // entry node --> (true branch, false branch)
                entry.setTrans(new Trans.Branch(entry, i.condition, tHead, fHead));

                if (fExit != null || tExit != null) {
                    exit = factory.mk();
                    // true branch --> end node.

                    if (tExit != null) {
                        tExit.setTrans(new Trans.Goto(tExit, exit));
                    }

                    // false branch --> end node.
                    if (fExit != null) {
                        fExit.setTrans(new Trans.Goto(fExit, exit));
                    }
                } else {
                    exit = null;
                }
            }

            return exit;
        }
        else if (tir instanceof TIRSwitch) {
            TIRSwitch s = (TIRSwitch) tir;

            @SuppressWarnings("unchecked")
            Pair<TIRConst.Value[],Node>[] tCases = new Pair[s.cases.length];

            ArrayList<Node> cExits = new ArrayList<Node>();

            for (int caseI = 0; caseI < tCases.length; caseI++) {
                TIRSwitch.Case c = s.cases[caseI];
                Node cHead = factory.mk();
                Node cExit = build(factory, c.body, cHead, blockMap);

                if (cExit != null) {
                    cExits.add(cExit);
                }

                tCases[caseI] = Pair.mk(c.value, cHead);
            }

            Node defTarget = null;
            if (s.defcase != null) {
                Node cHead = factory.mk();
                Node cExit = build(factory, s.defcase.body, cHead, blockMap);

                if (cExit != null) {
                    cExits.add(cExit);
                }

                defTarget = cHead;
            }

            entry.setTrans(new Trans.Switch(entry, s.expr, tCases, defTarget));

            if (!cExits.isEmpty()) {
                // If any of the cases had normal exits, then we need an overall "exit" block.
                Node exit = factory.mk();

                for (Node cExit : cExits) {
                    cExit.setTrans(new Trans.Goto(cExit, exit));
                }

                return exit;
            } else {
                return null;
            }
        }
        else if (tir instanceof TIRBlock) {
            TIRBlock b = (TIRBlock) tir;

            Node current;
            Node continueTarget;

            if (b.isLoop()) {
                // There's a "continue" in here somewhere, so we
                // need to entry a new block.
                current = factory.mk();
                continueTarget = current;
                entry.setTrans(new Trans.Goto(entry, current));
            } else {
                // No "continue", just reuse the incoming block.
                current = entry;
                continueTarget = null;
            }

            ArrayList<Node> breaks;
            if (b.hasBreak()) {
                breaks = new ArrayList<Node>();
            } else {
                breaks = null;
            }

            BlockMeta meta = new BlockMeta(continueTarget, breaks);
            Object displaced = blockMap.put(b, meta);
            assert displaced == null;

            for (TIRExpr e : b.list) {
                if (current == null) current = factory.mk();
                current = build(factory, e, current, blockMap);
            }

            Object removed = blockMap.remove(b);
            assert removed == meta;

            if (breaks != null) {
                Node after = factory.mk();
                for (Node breakSource : meta.breakSources) {
                    breakSource.setTrans(new Trans.Goto(breakSource, after));
                }
                if (current != null) {
                    current.setTrans(new Trans.Goto(current, after));
                }
                current = after;
            }

            return current;
        }
        else if (tir instanceof TIRBlock.Continue) {
            TIRBlock.Continue c = (TIRBlock.Continue) tir;
            BlockMeta meta = blockMap.get(c.block);
            assert meta != null;
            assert meta.continueTarget != null; // Only set if block is a loop block.
            entry.setTrans(new Trans.Goto(entry, meta.continueTarget));
            return null;
        }
        else if (tir instanceof TIRBlock.Break) {
            TIRBlock.Break c = (TIRBlock.Break) tir;
            BlockMeta meta = blockMap.get(c.block);
            assert meta != null;
            assert meta.breakSources != null; // Only set if block is a breakable block.
            meta.breakSources.add(entry);
            return null;
        }
        else if (tir instanceof TIRReturn) {
            TIRReturn r = (TIRReturn) tir;
            entry.setTrans(new Trans.Return(r.value));
            return null;
        }
        else if (tir instanceof TIRThrow) {
            TIRThrow t = (TIRThrow) tir;
            entry.setTrans(new Trans.Throw(t.exception));
            return null;
        }
        else if (tir instanceof TIROperator || tir instanceof TIRCall || tir instanceof TIRLocal.Set) {
            entry.stmts.add(tir);
            return entry;
        }
        else if (tir instanceof TIRConst.Value || tir instanceof TIRLocal.Get) {
            // Can safely ignore.
            return entry;
        }
        else {
            System.err.println("WARNING: Skipping @ " + tir.getSourcePoint() + ": " + tir.toString());
            return entry;
        }
    }

    public static final class Node
    {
        public int index;
        public final List<TIRExpr> stmts = new ArrayList<TIRExpr>();
        public Trans trans;
        public final List<Edge> incoming = new ArrayList<Edge>(2);

        public Node(int index)
        {
            this.index = index;
        }

        public void setTrans(Trans t)
        {
            assert this.trans == null;
            if (t instanceof Trans.NonExit) {
                assert ((Trans.NonExit)t).source == this;
            }
            this.trans = t;
            this.trans.connect();
        }

        public void clearTrans()
        {
            assert this.trans != null;
            this.trans.disconnect();
            this.trans = null;
        }

        public boolean equals(Object o) { return this == o; }
        public int hashCode() { return index; }
        public String toString() { return Integer.toString(index); }
    }

    public static abstract class Edge
    {
        private int index;
        public int getIndex() { return index; }
        void setIndex(int index) { this.index = index; }

        public final Node target;
        protected Edge(Node target)
        {
            this.target = target;
        }

        public abstract Trans.NonExit getTrans();
        public Node getSource() { return getTrans().source; }

        public static final class Branch extends Edge
        {
            public final Trans.Branch trans;
            public final boolean condition;

            public Branch(Trans.Branch trans, Node target, boolean condition)
            {
                super(target);
                this.trans = trans;
                this.condition = condition;
            }

            public boolean equals(Object o) { return getClass().equals(o.getClass()) && equals((Branch)o); }
            public boolean equals(Branch o) { return trans == o.trans && condition == o.condition; }
            public String toString() { return "Branch[" + trans.source.index + "](" + condition + ")"; }

            public Trans.Branch getTrans() { return trans; }
        }

        public static final class Goto extends Edge
        {
            public final Trans.Goto trans;

            public Goto(Trans.Goto trans, Node target)
            {
                super(target);
                this.trans = trans;
            }

            public boolean equals(Object o) { return getClass().equals(o.getClass()) && equals((Goto)o); }
            public boolean equals(Goto o) { return trans == o.trans; }
            public String toString() { return "Goto[" + trans.source.index + "]"; }

            public Trans.Goto getTrans() { return trans; }
        }

        public static abstract class Switch extends Edge
        {
            public final Trans.Switch trans;

            protected Switch(Trans.Switch trans, Node target)
            {
                super(target);
                this.trans = trans;
            }

            public Trans.Switch getTrans() { return trans; }

            public static final class Case extends Switch
            {
                public final int index;

                public Case(Trans.Switch trans, Node target, int index)
                {
                    super(trans, target);
                    this.index = index;
                }

                public boolean equals(Object o) { return getClass().equals(o.getClass()) && equals((Case)o); }
                public boolean equals(Case o) { return trans == o.trans && index == o.index; }
                public String toString() { return "Case[" + trans.source.index + "](" + index + ")"; }
            }

            public static final class Default extends Switch
            {
                public Default(Trans.Switch trans, Node target)
                {
                    super(trans, target);
                }

                public boolean equals(Object o) { return getClass().equals(o.getClass()) && equals((Default)o); }
                public boolean equals(Default o) { return trans == o.trans; }
                public String toString() { return "Default[" + trans.source.index + "]"; }
            }
        }
    }

    public static abstract class Trans
    {
        public abstract List<? extends Edge> getEdges();

        public void connect()
        {
            for (Edge e : getEdges()) {
                e.target.incoming.add(e);
            }
        }

        public void disconnect()
        {
            for (Edge e : getEdges()) {
                boolean removed = e.target.incoming.remove(e);
                assert removed;
            }
        }

        public abstract Trans changeSource(Node newSource);

        public static abstract class NonExit extends Trans
        {
            public final Node source;
            protected NonExit(Node source)
            {
                this.source = source;
            }
        }

        public static final class Branch extends NonExit
        {
            public final TIRExpr condition;
            public final Edge.Branch trueEdge;
            public final Edge.Branch falseEdge;

            public Branch(Node source, TIRExpr condition, Node trueTarget, Node falseTarget)
            {
                super(source);
                assert condition != null;
                assert trueTarget != null;
                assert falseTarget != null;
                this.condition = condition;
                this.trueEdge = new Edge.Branch(this, trueTarget, true);
                this.falseEdge = new Edge.Branch(this, falseTarget, false);
            }

            public String toString() { return "Branch(" + source + " -> " + trueEdge.target.index + "|" + falseEdge.target.index + ")"; }

            public List<Edge.Branch> getEdges()
            {
                return Arrays.asList(trueEdge, falseEdge);
            }

            public Branch changeTarget(Edge.Branch edge, Node newTarget)
            {
                assert edge.trans == this;
                if (edge.condition) {
                    return new Branch(source, condition, newTarget, falseEdge.target);
                } else {
                    return new Branch(source, condition, trueEdge.target, newTarget);
                }
            }

            @Override
            public Branch changeSource(Node newSource)
            {
                return new Branch(newSource, this.condition, trueEdge.target, falseEdge.target);
            }
        }

        public static final class Return extends Trans
        {
            public final TIRExpr value; // Might be null.

            public Return(TIRExpr value)
            {
                this.value = value;
            }

            public String toString() { return "Return"; }

            public List<Edge> getEdges()
            {
                return Collections.emptyList();
            }

            @Override
            public Return changeSource(Node newSource) { return this; }
        }

        public static final class Throw extends Trans
        {
            public final Class<? extends Operator.Exception> exception;

            public Throw(Class<? extends Operator.Exception> exception)
            {
                this.exception = exception;
            }

            public String toString() { return "Throw"; }

            public List<Edge> getEdges()
            {
                return Collections.emptyList();
            }

            @Override
            public Throw changeSource(Node newSource) { return this; }
        }

        public static final class Goto extends NonExit
        {
            public final Edge.Goto edge;

            public Goto(Node source, Node target)
            {
                super(source);
                assert target != null;
                this.edge = new Edge.Goto(this, target);
            }

            public String toString() { return "Goto(" + source + " -> " + edge.target.index + ")"; }

            public List<Edge.Goto> getEdges()
            {
                return Collections.singletonList(edge);
            }

            public Goto changeTarget(Node newTarget)
            {
                return new Goto(source, newTarget);
            }

            @Override
            public Goto changeSource(Node newSource)
            {
                return new Goto(newSource, edge.target);
            }
        }

        public static final class Switch extends NonExit
        {
            public final TIRExpr expr;
            public final Case[] cases;
            public final Edge.Switch.Default defEdge; // Might be null.

            public Switch(Node source, TIRExpr expr, Pair<TIRConst.Value[],Node>[] cases, Node defTarget)
            {
                super(source);
                this.expr = expr;
                if (defTarget != null) {
                    this.defEdge = new Edge.Switch.Default(this, defTarget);
                } else {
                    this.defEdge = null;
                }

                this.cases = new Case[cases.length];
                for (int caseI = 0; caseI < cases.length; caseI++) {
                    Pair<TIRConst.Value[],Node> c = cases[caseI];
                    this.cases[caseI] = new Case(c.left, new Edge.Switch.Case(this, c.right, caseI));
                }
            }

            public List<Edge.Switch> getEdges()
            {
                ArrayList<Edge.Switch> l = new ArrayList<Edge.Switch>();
                for (Case c : cases) {
                    l.add(c.edge);
                }
                if (defEdge != null) {
                    l.add(defEdge);
                }
                return l;
            }

            public static final class Case
            {
                public final TIRConst.Value[] values;
                public final Edge.Switch.Case edge;

                public Case(TIRConst.Value[] values, Edge.Switch.Case edge)
                {
                    this.values = values;
                    this.edge = edge;
                }
            }

            public Switch changeCaseTarget(Edge.Switch.Case edge, Node newTarget)
            {
                assert edge.trans == this;

                Pair<TIRConst.Value[],Node>[] newCases = copyCases();
                newCases[edge.index] = Pair.mk(this.cases[edge.index].values, newTarget);

                return new Switch(source, expr, newCases, defEdge.target);
            }

            public Switch changeDefaultTarget(Node newTarget)
            {
                return new Switch(source, expr, copyCases(), newTarget);
            }

            public Switch changeSource(Node newSource)
            {
                Node defTarget = (defEdge == null) ? null : defEdge.target;
                return new Switch(newSource, expr, copyCases(), defTarget);
            }

            private Pair<TIRConst.Value[],Node>[] copyCases()
            {
                @SuppressWarnings("unchecked")
                Pair<TIRConst.Value[],Node>[] newCases = new Pair[this.cases.length];
                for (int caseI = 0; caseI < newCases.length; caseI++) {
                    Case c = this.cases[caseI];
                    newCases[caseI] = Pair.mk(c.values, c.edge.target);
                }
                return newCases;
            }
        }
    }

    public static final class NodeSet implements Iterable<Node>
    {
        private final int[] bits;
        private Link list = null;

        public NodeSet(Cfg cfg)
        {
            int nBits = cfg.nodes.length;
            int nInts = (nBits+31) / 32;
            bits = new int[nInts];
        }

        private NodeSet(int[] bits, Link head)
        {
            this.bits = bits;
            this.list = head;
        }

        public boolean add(Node n)
        {
            int index = n.index;
            int intOffset = index / 32;
            int bitOffset = index % 32;
            int mask = 1 << bitOffset;

            int val = bits[intOffset];
            if ((val & mask) == 0) {
                this.list = new Link(n, this.list);
                bits[intOffset] = val | mask;
                return true;
            }
            else {
                return false;
            }
        }

        public boolean contains(Node n)
        {
            int index = n.index;
            int intOffset = index / 32;
            int bitOffset = index % 32;
            int mask = 1 << bitOffset;

            int val = bits[intOffset];
            return ((val & mask) != 0);
        }

        public NodeSet copy()
        {
            int[] bits = new int[this.bits.length];
            System.arraycopy(this.bits, 0, bits, 0, bits.length);
            return new NodeSet(bits, this.list);
        }

        public Node removeLast()
        {
            assert !isEmpty();
            Link head = list;
            list = list.next;

            Node n = head.node;
            int index = n.index;
            int intOffset = index / 32;
            int bitOffset = index % 32;
            int mask = 1 << bitOffset;

            int val = bits[intOffset];
            assert (val & mask) != 0;
            val &= ~mask; // Unset the bit.

            return n;
        }

        public boolean isEmpty()
        {
            return list == null;
        }

        public java.util.Iterator<Node> iterator()
        {
            return new Iterator(this.list);
        }

        private static final class Link
        {
            public final Node node;
            public final Link next;

            private Link(Node node, Link next)
            {
                this.node = node;
                this.next = next;
            }
        }

        private static final class Iterator implements java.util.Iterator<Node>
        {
            private Link current;

            private Iterator(Link current)
            {
                this.current = current;
            }

            public boolean hasNext() { return current != null; }

            public Node next()
            {
                if (current == null) throw new java.util.NoSuchElementException();
                Node n = current.node;
                current = current.next;
                return n;
            }

            public void remove()
            {
                throw new UnsupportedOperationException();
            }
        }
    }
}
