package vpc.cfg;

import vpc.tir.*;
import vpc.tir.expr.Operator;
import vpc.hil.Device;
import vpc.core.base.PrimInt32;
import vpc.core.base.PrimChar;
import vpc.core.base.PrimBool;
import vpc.core.base.PrimRaw;
import vpc.core.base.PrimConversion;
import vpc.core.base.Reference;
import vpc.core.virgil.VirgilClass;
import vpc.core.virgil.VirgilComponent;
import vpc.core.virgil.VirgilArray;
import vpc.core.Value;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;

public class CfgNodeRenderer
{
    public static final CfgNodeRenderer Instance = new CfgNodeRenderer();

    public Iterable<String> before(Cfg.Node n)
    {
        return Collections.emptyList();
    }

    public Iterable<String> after(Cfg.Node n)
    {
        return Collections.emptyList();
    }

    public final void render(StringBuilder buf, TIRExpr e)
    {
        render(buf, e, false);
    }

    public void render(StringBuilder buf, TIRExpr e, boolean needParens)
    {
        if (e instanceof TIRLocal.Get) {
            TIRLocal.Get g = (TIRLocal.Get) e;
            buf.append(g.temp.getName());
        }
        else if (e instanceof TIRLocal.Set) {
            TIRLocal.Set s = (TIRLocal.Set) e;
            buf.append(s.temp.getName());
            buf.append(" <- ");
            render(buf, s.value);
        }
        else if (e instanceof TIRCall) {
            TIRCall c = (TIRCall) e;
            render(buf, c.func);
            buf.append("(");
            String sep = "";
            for (TIRExpr arg : c.arguments) {
                buf.append(sep); sep = ", ";
                render(buf, arg);
            }
            buf.append(")");
        }
        else if (e instanceof TIRConst.Symbol) {
            TIRConst.Symbol s = (TIRConst.Symbol) e;
            buf.append(s.value());
        }
        else if (e instanceof TIRConst.Value) {
            TIRConst.Value v = (TIRConst.Value) e;
            buf.append(v.getValue());
        }
        else if (e instanceof TIROperator) {
            TIROperator op = (TIROperator) e;
            Operator func = op.operator;
            if (func instanceof VirgilClass.GetField) {
                VirgilClass.GetField gf = (VirgilClass.GetField) func;
                render(buf, op.operands[0]);
                buf.append(".");
                buf.append(gf.field.getName());
            }
            else if (func instanceof VirgilClass.SetField) {
                VirgilClass.SetField sf = (VirgilClass.SetField) func;
                render(buf, op.operands[0]);
                buf.append(".");
                buf.append(sf.field.getName());
                buf.append(" <- ");
                render(buf, op.operands[1]);
            }
            else if (func instanceof VirgilComponent.GetField) {
                VirgilComponent.GetField gf = (VirgilComponent.GetField) func;
                buf.append(gf.component.getName());
                buf.append(".");
                buf.append(gf.field.getName());
            }
            else if (func instanceof VirgilComponent.SetField) {
                VirgilComponent.SetField sf = (VirgilComponent.SetField) func;
                buf.append(sf.component.getName());
                buf.append(".");
                buf.append(sf.field.getName());
                buf.append(" <- ");
                render(buf, op.operands[0]);
            }
            else if (func instanceof VirgilArray.GetElement) {
                render(buf, op.operands[0]);
                buf.append('[');
                render(buf, op.operands[1]);
                buf.append(']');
            }
            else if (func instanceof VirgilArray.SetElement) {
                render(buf, op.operands[0]);
                buf.append('[');
                render(buf, op.operands[1]);
                buf.append("] <- ");
                render(buf, op.operands[2]);
            }
            else if (func instanceof Reference.NullCheck) {
                buf.append("Null?(");
                render(buf, op.operands[0]);
                buf.append(")");
            }
            else if (func instanceof VirgilClass.GetMethod) {
                VirgilClass.GetMethod gm = (VirgilClass.GetMethod) func;
                render(buf, op.operands[0]);
                buf.append(".");
                buf.append(gm.method.getName());
            }
            else if (func instanceof VirgilComponent.GetMethod) {
                VirgilComponent.GetMethod gm = (VirgilComponent.GetMethod) func;
                buf.append(gm.component.getName());
                buf.append(".");
                buf.append(gm.method.getName());
            }
            else if (BinOpSymbols.containsKey(func.getClass())) {
                String symbol = BinOpSymbols.get(func.getClass());
                open(buf, needParens);
                render(buf, op.operands[0]);
                buf.append(' ').append(symbol).append(' ');
                render(buf, op.operands[1]);
                close(buf, needParens);
            }
            else if (UnOpSymbols.containsKey(func.getClass())) {
                String symbol = UnOpSymbols.get(func.getClass());
                open(buf, needParens);
                buf.append(symbol).append(' ');
                render(buf, op.operands[0], true);
                close(buf, needParens);
            }
            else if (func instanceof Device.GetRegister) {
                Device.GetRegister gr = (Device.GetRegister) func;
                buf.append(gr.register.regionToken.image + "[" + gr.register.index.image + "]");
            }
            else if (func instanceof Device.SetRegister) {
                Device.SetRegister sr = (Device.SetRegister) func;
                buf.append(sr.register.regionToken.image + "[" + sr.register.index.image + "] <- ");
                render(buf, op.operands[0]);
            }
            else {
                String funcName = FuncNames.get(func.getClass());
                if (funcName == null) funcName = makeFuncName(func);

                buf.append(funcName);
                buf.append('(');
                String sep = "";
                for (TIRExpr operand : op.operands) {
                    buf.append(sep); sep = ", ";
                    render(buf, operand);
                }
                buf.append(')');
            }
        }
        else {
            // Use TIRExpr.toString() if we don't have a special handler.
            buf.append(e.toString());
        }
    }

    private static String makeFuncName(Operator op)
    {
        // Get the class name without the package prefix.
        Class<?> c = op.getClass();
        String name = c.getSimpleName();

        // Add in the enclosing class names.
        while ((c = c.getEnclosingClass()) != null) {
            name = c.getSimpleName() + "." + name;
        }

        return "$" + name;
    }

    private static void open(StringBuilder buf, boolean needParens)
    {
        if (needParens) buf.append('(');
    }


    private static void close(StringBuilder buf, boolean needParens)
    {
        if (needParens) buf.append(')');
    }

    private static final Map<Class<? extends Operator.Op2>,String> BinOpSymbols = new HashMap<Class<? extends Operator.Op2>,String>();
    static {
        Map<Class<? extends Operator.Op2>,String> m = BinOpSymbols;
        m.put(PrimInt32.ADD.class, "+");
        m.put(PrimInt32.SUB.class, "-");
        m.put(PrimInt32.MUL.class, "*");
        m.put(PrimInt32.DIV.class, "/");
        m.put(PrimInt32.MOD.class, "%");
        m.put(PrimInt32.AND.class, "&");
        m.put(PrimInt32.OR.class, "|");
        m.put(PrimInt32.XOR.class, "^");
        m.put(PrimInt32.EQU.class, "=");
        m.put(PrimInt32.NEQU.class, "!=");
        m.put(PrimInt32.GR.class, ">");
        m.put(PrimInt32.GREQ.class, ">=");
        m.put(PrimInt32.LT.class, "<");
        m.put(PrimInt32.LTEQ.class, "<=");
        m.put(PrimInt32.SHL.class, "<<");
        m.put(PrimInt32.SHR.class, ">>");
        m.put(PrimChar.EQU.class, "=");
        m.put(PrimChar.NEQU.class, "!=");
        m.put(PrimChar.GR.class, ">");
        m.put(PrimChar.GREQ.class, ">=");
        m.put(PrimChar.LT.class, "<");
        m.put(PrimChar.LTEQ.class, "<=");
        m.put(PrimBool.AND.class, "&&");
        m.put(PrimBool.OR.class, "||");
        m.put(PrimRaw.AND.class, "&");
        m.put(PrimRaw.OR.class, "|");
        m.put(PrimRaw.XOR.class, "^");
        m.put(PrimRaw.SHL.class, "<<");
        m.put(PrimRaw.SHR.class, ">>");
        m.put(Value.Equal.class, "=");
        m.put(Value.NotEqual.class, "!=");
    }

    private static final Map<Class<? extends Operator.Op1>,String> UnOpSymbols = new HashMap<Class<? extends Operator.Op1>,String>();
    static {
        Map<Class<? extends Operator.Op1>,String> m = UnOpSymbols;
        m.put(PrimInt32.NEG.class, "-");
        m.put(PrimBool.NOT.class, "!");
        m.put(PrimRaw.Complement.class, "~");
        m.put(PrimConversion.Char_Int32.class, "{Char->Int32}");
        m.put(PrimConversion.Char_Raw.class, "{Char->Raw}");
        m.put(PrimConversion.Int32_Char.class, "{Int32->Char}");
        m.put(PrimConversion.Int32_Raw.class, "{Int32->Raw}");
        m.put(PrimConversion.Raw_Char.class, "{Raw->Char}");
        m.put(PrimConversion.Raw_Int32.class, "{Raw->Int32}");
    }

    private static final Map<Class<? extends Operator>,String> FuncNames = new HashMap<Class<? extends Operator>,String>();
    static {
        Map<Class<? extends Operator>,String> m = FuncNames;
        m.put(VirgilArray.GetLength.class, "$ArrayLength");
        m.put(PrimRaw.SetBit.class, "$SetBit");
        m.put(PrimRaw.GetBit.class, "$GetBit");
        m.put(PrimRaw.Concat.class, "$Concat");
        m.put(PrimConversion.AdjustRaw.class, "$AdjustRaw");
        m.put(VirgilArray.Alloc.class, "$ArrayAlloc");
    }
}
