/**
 * Copyright (c) 2006, Regents of the University of California
 * See the file "license.txt" for details.
 *
 * Creation date: Mar 24, 2006
 */

package vpc.core.base;

import cck.util.Util;
import vpc.core.Value;
import vpc.core.types.*;
import vpc.tir.expr.Operator;
import vpc.tir.expr.Precedence;

/**
 * The <code>PrimByte</code> class represents a byte (an 8-bit unsigned integer).
 *
 * @author Ben L. Titzer
 */
public class PrimByte {

    public static Type.Simple TYPE = new Type.Simple("byte");
    public static TypeCon TYPECON = new TypeCon.Singleton(TYPE);

    static {
        // add the binops to each of the primitive types
        Precedence.register(EQU.class, Precedence.PREC_EQUAL);
        Precedence.register(NEQU.class, Precedence.PREC_EQUAL);
        Precedence.register(LT.class, Precedence.PREC_LESS_THAN);
        Precedence.register(GR.class, Precedence.PREC_GREATER);
        Precedence.register(LTEQ.class, Precedence.PREC_LESS_EQUAL);
        Precedence.register(GREQ.class, Precedence.PREC_GREATER_EQUAL);
    }

    public abstract static class Compare extends Operator.Op2 {

        protected Compare() {
            super(TYPE, TYPE, PrimBool.TYPE);
        }

        public Value apply2(Value v1, Value v2) {
            return PrimBool.toValue(apply(fromValue(v1), fromValue(v2)));
        }

        public abstract boolean apply(int v1, int v2);

    }

    /**
     * The <code>Comparison.PrimByte.EQU</code> class represents an integer comparison for equality.
     */
    public static class EQU extends Compare {
        public boolean apply(int v1, int v2) {
            return v1 == v2;
        }
    }

    /**
     * The <code>Comparison.PrimByte.NEQU</code> class represents an integer comparison for inequality.
     */
    public static class NEQU extends Compare {
        public boolean apply(int v1, int v2) {
            return v1 != v2;
        }
    }

    /**
     * The <code>Comparison.PrimByte.GR</code> class represents an integer comparison for greater than.
     */
    public static class GR extends Compare {
        public boolean apply(int v1, int v2) {
            return v1 > v2;
        }
    }

    /**
     * The <code>Comparison.PrimByte.LT</code> class represents an integer comparison for less than.
     */
    public static class LT extends Compare {
        public boolean apply(int v1, int v2) {
            return v1 < v2;
        }
    }

    /**
     * The <code>Comparison.PrimByte.GREQ</code> class represents an integer comparison
     * for greater than or equal to.
     */
    public static class GREQ extends Compare {
        public boolean apply(int v1, int v2) {
            return v1 >= v2;
        }
    }

    /**
     * The <code>Comparison.PrimByte.LTEQ</code> class represents an integer comparison for less than
     * or equal to.
     */
    public static class LTEQ extends Compare {
        public boolean apply(int v1, int v2) {
            return v1 <= v2;
        }
    }

    /**
     * The implementation of byte values.
     */
    public static class Val extends Value {
        public final int value;
        public final PrimInt32.Val intVal; // cache of integer val for performance

        private static Val[] cache = new Val[256];

        static {
            // all byte values are cached
            for (int i = 0; i < 256; i++) {
                cache[i] = new Val(i);
            }
        }

        private Val(int v) {
            super();
            value = v;
            intVal = new PrimInt32.Val(value);
        }

        public Type getType() {
            return TYPE;
        }

        public boolean equals(Object o) {
            if (o == this) return true;
            if (o == Value.BOTTOM ) return value == 0; // 0 == bottom
            return o instanceof Val && ((Val) o).value == value;
        }

        public String toString() {
            return "byte:" + (int) value;
        }
    }

    public static int fromValue(Value v) {
        if (v == Value.BOTTOM) return (char)0; // null (bottom) means the zero char
        if (v instanceof Val) return ((Val) v).value;
        throw Util.failure("Value of type " + v.getType() + " cannot be converted to PrimByte");
    }

    public static Val toValue(int c) {
        return Val.cache[c & 0xff];
    }
}