package vpc.ssa;

import vpc.cfg.Cfg;

import java.util.Iterator;

// Calculate dominator relation for a CFG
//
// Dominance and dominance frontier algorithms taken from:
//    A Simple, Fast Dominance Algorithm (2001)
//    Keith D Cooper, Timothy J Harvey, Ken Kennedy
//    Software Practice and Experience
public class Dom
{
    private Dom() {}

    public static int[] calcIdoms(Cfg cfg)
    {
        int n = cfg.nodes.length;
        int[] idoms = new int[n];

        if (cfg.entry.index != 0) throw new AssertionError("The entry node must have index zero.  cfg.entry.index = " + cfg.entry.index);

        // Initialize everything to "undefined", except the entry node.
        for (int i = 1; i < n; i++) {
            idoms[i] = -1;
        }

        while (true) {
            boolean changed = false;

            // The CFG builder orders the nodes in a way that is good
            // for this algorithm (predecessors before successors).  So
            // we can just iterate through directly.
            for (int i = 1; i < n; i++) {
                Cfg.Node b = cfg.nodes[i];

                // Find the first predecessor with an IDOM set.
                Iterator<Cfg.Edge> predi = b.incoming.iterator();
                while (predi.hasNext()) {
                    int firstPred = predi.next().getTrans().source.index;
                    if (idoms[firstPred] != -1) {
                        
                        // Intersect with every other predecessor with an IDOM set.
                        int idom = firstPred;
                        while (predi.hasNext()) {
                            int nextPred = predi.next().getTrans().source.index ;
                            if (idoms[nextPred] != -1) {
                                idom = intersect(idoms, idom, nextPred);
                            }
                        }

                        // If the IDOM we calculated is different from the old one...
                        if (idoms[b.index] != idom) {
                            idoms[b.index] = idom;
                            changed = true;
                        }

                        break;
                    }
                }

            }

            if (!changed) break;
        }

        return idoms;
    }

    private static int intersect(int[] idoms, int idom, int predIdom)
    {
        assert idom != -1;
        assert predIdom != -1;

        int finger1 = idom;
        int finger2 = predIdom;
        while (finger1 != finger2) {
            while (finger1 > finger2) {
                finger1 = idoms[finger1];
                assert finger1 != -1;
            }
            while (finger2 > finger1) {
                finger2 = idoms[finger2];
                assert finger2 != -1;
            }
        }

        return finger1;
    }

    // Since Virgil functions have tiny CFGs, we can represent sets using a matrix.
    public static Cfg.NodeSet[] calcFrontiers(Cfg cfg, int[] idoms)
    {
        assert cfg.nodes.length == idoms.length;
        int len = idoms.length;

        Cfg.NodeSet[] dfs = new Cfg.NodeSet[len];
        for (int i = 0; i < dfs.length; i++) {
            dfs[i] = new Cfg.NodeSet(cfg);
        }

        for (int i = 1; i < len; i++) {
            Cfg.Node n = cfg.nodes[i];
            int ni = n.index;
            int numPreds = n.incoming.size();
            if (n == cfg.entry) numPreds++;
            if (numPreds >= 2) {
                for (Cfg.Edge rt : n.incoming) {
                    int runner = rt.getTrans().source.index;
                    while (runner != idoms[ni]) {
                        dfs[runner].add(n);
                        runner = idoms[runner];
                    }
                }
            }
        }

        return dfs;
    }
}
