package vpc.ssa;

import vpc.tir.*;
import vpc.cfg.Cfg;
import vpc.cfg.Cfg.Node;
import vpc.cfg.Cfg.NodeSet;
import vpc.cfg.Cfg.Trans;
import vpc.cfg.Cfg.Edge;
import vpc.core.decl.Method.Temporary;
import vpc.core.types.Type;
import vpc.util.IntStack;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;
import java.util.Collection;
import java.util.HashSet;
import java.util.Queue;
import java.util.LinkedList;
import java.util.Set;

public class Ssa
{
    public final Phi[][] phis;
    public final Def.Param[] params;
    public final Map<TIRLocal.Set,Def.Set> sets;
    public final Map<TIRLocal.Get,Use.Get> gets;

    private Ssa(int numNodes, int numParams)
    {
        this.phis = new Phi[numNodes][];
        this.params = new Def.Param[numParams];
        this.sets = new HashMap<TIRLocal.Set,Def.Set>();
        this.gets = new HashMap<TIRLocal.Get,Use.Get>();
    }

    // ---------------------------------------------------------------------
    // Data Structures
    // ---------------------------------------------------------------------

    public static final class Var
    {
        public final Temporary<Type> base;
        public final int number;
        public final Def def;
        public final List<Use> uses = new ArrayList<Use>();

        public Var(Temporary<Type> base, int number, Def def)
        {
            this.base = base;
            this.number = number;
            this.def = def;
        }

        public final int hashCode() { return base.hashCode() + number; }
        public final boolean equals(Object o) { return this == o; }
    }

    public static final class Phi extends Def
    {
        public final Var[] args;

        public Phi(int numArgs)
        {
            this.args = new Var[numArgs];
        }
    }

    public static abstract class Ref // Use or Def
    {
        private Var var;

        public Var getVar()
        {
            assert var != null;
            return var;
        }

        void setVar(Var var)
        {
            assert var != null;
            assert this.var == null;
            this.var = var;
        }

    }

    public static abstract class Def extends Ref
    {
        public static abstract class Normal extends Def
        {
        }

        public static final class Set extends Normal
        {
            public final TIRLocal.Set set;

            public Set(TIRLocal.Set set)
            {
                this.set = set;
            }
        }

        public static final class Param extends Normal
        {
            public final int index;

            public Param(int index)
            {
                this.index = index;
            }
        }

        public static final class Undefined extends Def
        {
        }
    }

    public static abstract class Use extends Ref
    {
        public static abstract class Normal extends Use
        {
        }

        public static final class Get extends Normal
        {
            public final TIRLocal.Get get;

            public Get(TIRLocal.Get get)
            {
                this.get = get;
            }
        }

        public static final class Phi extends Use
        {
            public final Ssa.Phi phi;
            public final int index;

            public Phi(Ssa.Phi phi, int index)
            {
                this.phi = phi;
                this.index = index;
            }
        }
    }

    // ---------------------------------------------------------------------
    // SSA Construction
    // ---------------------------------------------------------------------

    public static Ssa build(TIRRep rep, Cfg cfg)
    {
        int[] idoms = Dom.calcIdoms(cfg);
        NodeSet[] dfs = Dom.calcFrontiers(cfg, idoms);

        int numNodes = cfg.nodes.length;

        int numParams = countIterable(rep.getParams());
        Ssa ssa = new Ssa(numNodes, numParams);

        int numVars = rep.getNumberOfTemps();

        NodeSet[] defSites = new NodeSet[numVars];
        for (int i = 0; i < defSites.length; i++) {
            defSites[i] = new NodeSet(cfg);
        }

        // Find uses and defs.
        @SuppressWarnings("unchecked")
        List<RefInfo>[] refs = new List[numNodes];
        HashSet<Temporary<Type>> liveIns = new HashSet<Temporary<Type>>();
        for (Node n : cfg.nodes) {
            refs[n.index] = findRefs(ssa, n, defSites, liveIns);
        }

        // Place PHIs
        HashMap<Temporary<Type>,Phi>[] phis = placePhis(cfg, defSites, dfs, rep.getTemps(), liveIns);

        // Rename variables.
        rename(cfg, ssa, numVars, rep.getParams(), phis, refs);

        // Prune PHIs
        prune(phis, refs);

        for (int i = 0; i < numNodes; i++) {
            Collection<Phi> nodePhis = phis[i].values();
            ssa.phis[i] = nodePhis.toArray(new Phi[nodePhis.size()]);
        }

        return ssa;
    }

    private static int countIterable(Iterable<?> it)
    {
        int count = 0;
        Iterator<?> iterator = it.iterator();
        while (iterator.hasNext()) {
            iterator.next();
            count++;
        }
        return count;
    }

    private static HashMap<Temporary<Type>,Phi>[] placePhis(Cfg cfg, NodeSet[] defSites, NodeSet[] dfs, Iterable<Temporary<Type>> vars, HashSet<Temporary<Type>> liveIns)
    {
        int numNodes = cfg.nodes.length;

        @SuppressWarnings("unchecked")
        HashMap<Temporary<Type>,Phi>[] phis = new HashMap[numNodes]; // Remember where we inserted PHIs.

        for (int i = 0; i < numNodes; i++) {
            phis[i] = new HashMap<Temporary<Type>,Phi>();
        }

        // Place PHIs on dominance frontiers.
        for (Temporary<Type> var : vars) {
            // If the variable isn't live upon entry into any block, then it will
            // never need a PHI.  This is the "semi-pruned SSA" idea from one of
            // the SUIF guys.  A quick way of reducing the number of PHIs inserted.
            if (!liveIns.contains(var)) continue;

            NodeSet todo = defSites[var.id].copy();
            while (!todo.isEmpty()) {
                Node n = todo.removeLast();
                for (Node y : dfs[n.index]) {
                    assert y != cfg.entry;
                    // Add a PHI for variable 'var' in block 'y'.
                    if (!phis[y.index].containsKey(var)) {
                        int numPreds = y.incoming.size();
                        phis[y.index].put(var, new Phi(numPreds));
                        if (!defSites[var.id].contains(y)) {
                            // The added PHI means that 'y' now has a definition for 'var'.
                            // So we need to add it to the work list.
                            todo.add(y);
                        }
                    }
                }
            }
        }

        return phis;
    }

    // -----------------------------------------
    // Rename Variables

    private static void rename(Cfg cfg, Ssa ssa, int numVars, Iterable<Temporary<Type>> params, HashMap<Temporary<Type>,Phi>[] phis, List<RefInfo>[] refs)
    {
        VarStates varStates = new VarStates(numVars);

        // Handle function parameters first.
        int paramI = 0;
        for (Temporary<Type> param : params) {
            Def.Param d = new Def.Param(paramI);
            ssa.params[paramI] = d;
            varStates.def(d, param);
            paramI++;
        }

        boolean[] visited = new boolean[cfg.nodes.length];
        rename(varStates, phis, refs, visited, cfg.entry);
    }

    private static void rename(VarStates varStates, HashMap<Temporary<Type>,Phi>[] phis, List<RefInfo>[] refs, boolean[] visited, Node n)
    {
        if (visited[n.index]) return;
        visited[n.index] = true;

        varStates.pushState();

        // The PHIs are logically at the top of each CFG node.  So process them
        // first.
        HashMap<Temporary<Type>,Phi> nodePhis = phis[n.index];
        for (Map.Entry<Temporary<Type>,Phi> entry : nodePhis.entrySet()) {
            Temporary<Type> baseVar = entry.getKey();
            Phi phi = entry.getValue();
            varStates.def(phi, baseVar);
        }

        // Process all the uses and defs in the CFG node.
        List<RefInfo> nodeRefs = refs[n.index];
        for (RefInfo ri : nodeRefs) {
            Ref r = ri.ref;
            if (r instanceof Def) {
                Def d = (Def) r;
                varStates.def(d, ri.var);
            }
            else if (r instanceof Use) {
                Use u = (Use) r;
                varStates.use(u, ri.var);
            }
            else {
                throw new AssertionError("bad type: " + r.getClass().getName());
            }
        }

        // Peek into successor and set PHI args.
        for (Edge e : n.trans.getEdges()) {
            HashMap<Temporary<Type>,Phi> succPhis = phis[e.target.index];
            for (Map.Entry<Temporary<Type>,Phi> entry : succPhis.entrySet()) {
                Temporary<Type> baseVar = entry.getKey();
                Phi phi = entry.getValue();

                int argIndex = e.getIndex();
                Use.Phi use = new Use.Phi(phi, argIndex);
                varStates.use(use, baseVar);
                phi.args[e.getIndex()] = use.getVar();
            }

            // Appel says to visit nodes in dominator tree depth-first order.  But I think
            // plain depth-first order should work as well (plus, it's easier).
            rename(varStates, phis, refs, visited, e.target);
        }

        varStates.popState();
    }

    private static final class VarStates
    {
        private Entry[] entries;

        // To undo pushes when we backtrack in DFS.
        private IntStack pushes = new IntStack();
        private IntStack numPushesHistory = new IntStack();
        private int numPushes = 0;

        public VarStates(int numVars)
        {
            entries = new Entry[numVars];
            for (int i = 0; i < numVars; i++) {
                entries[i] = new Entry();
            }
        }

        public void pushState()
        {
            numPushesHistory.push(numPushes);
            numPushes = 0;
        }

        public void popState()
        {
            for (int i = 0; i < numPushes; i++) {
                int varId = pushes.pop();
                entries[varId].stack.pop();
            }
            numPushes = numPushesHistory.pop();
        }

        public Var def(Def def, Temporary<Type> var)
        {
            Entry e = entries[var.id];
            Var ssaVar = new Var(var, e.nextIndex++, def);
            e.stack.push(ssaVar);

            // Remember to undo the "e.stack.push" on the next popState().
            pushes.push(var.id);
            numPushes++;

            def.setVar(ssaVar);

            return ssaVar;
        }

        public Var use(Use use, Temporary<Type> var)
        {
            Entry e = entries[var.id];

            Var ssaVar;
            if (e.stack.isEmpty()) {
                if (e.undefined == null) {
                    Def.Undefined d = new Def.Undefined();
                    e.undefined = new Var(var, e.nextIndex++, d);
                    d.setVar(e.undefined);
                }
                ssaVar = e.undefined;
            }
            else {
                ssaVar = e.stack.peek();
            }

            use.setVar(ssaVar);
            ssaVar.uses.add(use);

            return ssaVar;
        }

        private static final class Entry
        {
            public int nextIndex = 0;
            public Stack<Var> stack = new Stack<Var>();
            public Var undefined = null;
        }
    }

    // -----------------------------------------
    // Locate uses and defs in a CFG node.

    private static List<RefInfo> findRefs(Ssa ssa, Node n, NodeSet[] defSites, HashSet<Temporary<Type>> liveIns)
    {
        ArrayList<RefInfo> refs = new ArrayList<RefInfo>();
        RefTracker tracker = new RefTracker(refs, n, defSites);

        // Look through the block's statements.
        for (TIRExpr s : n.stmts) {
            findRefs(ssa, tracker, s);
        }

        // Look through any TIR expressions in the transitions.
        Trans t = n.trans;
        if (t instanceof Trans.Branch) {
            Trans.Branch b = (Trans.Branch) t;
            findRefs(ssa, tracker, b.condition);
        }
        else if (t instanceof Trans.Switch) {
            Trans.Switch sw = (Trans.Switch) t;
            findRefs(ssa, tracker, sw.expr);
            // Don't have to visit the cases because their expressions
            // are all constant values (no var uses).
        }
        else if (t instanceof Trans.Goto) {
            // No TIR to visit.
        }
        else if (t instanceof Trans.Return) {
            Trans.Return r = (Trans.Return) t;
            if (r.value != null) findRefs(ssa, tracker, r.value);
        }
        else if (t instanceof Trans.Throw) {
            // No TIR to visit.
        }
        else {
            throw new AssertionError("bad type: " + t.getClass().getName());
        }

        liveIns.addAll(tracker.usedBeforeDefd);
        return refs;
    }

    private static void findRefs(Ssa ssa, RefTracker tracker, TIRExpr e)
    {
        if (e instanceof TIRLocal.Get) {
            TIRLocal.Get g = (TIRLocal.Get) e;
            Use.Get u = new Use.Get(g);
            tracker.use(g.temp, u);
            ssa.gets.put(g, u);
        }
        else if (e instanceof TIRLocal.Set) {
            TIRLocal.Set s = (TIRLocal.Set) e;
            // First explore the right-hand side, since that executes first.
            findRefs(ssa, tracker, s.value);
            // Then record the def.
            Def.Set d = new Def.Set(s);
            tracker.def(s.temp, d);
            ssa.sets.put(s, d);
        }
        else if (e instanceof TIRCall) {
            TIRCall c = (TIRCall) e;
            findRefs(ssa, tracker, c.func);
            findRefs(ssa, tracker, c.arguments);
        }
        else if (e instanceof TIROperator) {
            TIROperator op = (TIROperator) e;
            findRefs(ssa, tracker, op.operands);
        }
        else if (e instanceof TIRConst.Value || e instanceof TIRConst.Symbol) {
            // Nothing to see here.
        }
        else {
            throw new AssertionError("bad type: " + e.getClass().getName());
        }
    }

    private static void findRefs(Ssa ssa, RefTracker tracker, TIRExpr[] es)
    {
        for (TIRExpr e : es) {
            findRefs(ssa, tracker, e);
        }
    }

    private static final class RefTracker
    {
        public final List<RefInfo> refs;
        public final Node currentNode;
        public final NodeSet[] defSites;
        public final HashSet<Temporary<Type>> defd = new HashSet<Temporary<Type>>();
        public final HashSet<Temporary<Type>> usedBeforeDefd = new HashSet<Temporary<Type>>();

        public RefTracker(List<RefInfo> refs, Node currentNode, NodeSet[] defSites)
        {
            this.refs = refs;
            this.currentNode = currentNode;
            this.defSites = defSites;
        }

        public void use(Temporary<Type> var, Use.Normal u)
        {
            refs.add(new RefInfo(var, u));
            if (!defd.contains(var)) {
                usedBeforeDefd.add(var);
            }
        }

        public void def(Temporary<Type> var, Def.Normal d)
        {
            refs.add(new RefInfo(var, d));
            defSites[var.id].add(currentNode);
            defd.add(var);
        }
    }

    private static final class RefInfo
    {
        public final Temporary<Type> var;
        public final Ref ref;

        private RefInfo(Temporary<Type> var, Ref ref)
        {
            this.var = var;
            this.ref = ref;
        }
    }

    // -----------------------------------------
    // Prune

    private static void prune(HashMap<Temporary<Type>,Phi>[] phis, List<RefInfo>[] refs)
    {
        // Reachability problem.

        WorkQueue<Phi> queue = new WorkQueue<Phi>();

        // 1. The reachability roots are all non-PHI uses.
        for (List<RefInfo> nodeRefs : refs) {
            for (RefInfo ri : nodeRefs) {
                if (ri.ref instanceof Use) {
                    assert ri.ref instanceof Use.Normal;
                    Use.Normal use = (Use.Normal) ri.ref;
                    Def d = use.getVar().def;
                    if (d instanceof Phi) {
                        queue.add((Phi) d);
                    }
                }
            }
        }

        // 2. Compute the closure.
        while (!queue.isEmpty()) {
            Phi phi = queue.next();
            for (Var arg : phi.args) {
                Def d = arg.def;
                if (d instanceof Phi) {
                    queue.add((Phi) d);
                }
            }
        }

        // 3. Remove PHIs that aren't part of the closure.
        for (HashMap<Temporary<Type>,Phi> nodePhis : phis) {
            Iterator<Phi> it = nodePhis.values().iterator();
            while (it.hasNext()) {
                Phi phi = it.next();
                if (!queue.reached(phi)) {
                    it.remove();
                }
            }
        }
    }

    private static final class WorkQueue<T>
    {
        private final Set<T> reached = new HashSet<T>();
        private final Queue<T> todo = new LinkedList<T>();

        public boolean isEmpty()
        {
            return todo.isEmpty();
        }

        public T next()
        {
            if (isEmpty()) throw new java.util.NoSuchElementException();
            return todo.remove();
        }

        public void add(T t)
        {
            boolean added = reached.add(t);
            if (added) todo.add(t);
        }

        public boolean reached(T t)
        {
            return reached.contains(t);
        }
    }

}
